use crate::{platform, script_bridge, Layout};
use std::cell::Cell;
use std::mem;
use std::ops::Not;
use std::rc::Rc;
use wellen::GetItem;
use zoon::*;

const SCOPE_VAR_ROW_MAX_WIDTH: u32 = 480;
const MILLER_COLUMN_MAX_HEIGHT: u32 = 500;

type Filename = String;

#[derive(Clone)]
struct VarForUI {
    name: Rc<String>,
    var_type: wellen::VarType,
    var_direction: wellen::VarDirection,
    var_ref: wellen::VarRef,
    signal_type: wellen::SignalType,
}

#[derive(Clone)]
struct ScopeForUI {
    level: usize,
    name: Rc<String>,
    scope_ref: wellen::ScopeRef,
    has_children: bool,
    expanded: Mutable<bool>,
    parent_expanded: Option<ReadOnlyMutable<bool>>,
    selected_scope_in_level: Mutable<Option<wellen::ScopeRef>>,
}

#[derive(Clone)]
pub struct ControlsPanel {
    selected_scope_ref: Mutable<Option<wellen::ScopeRef>>,
    hierarchy: Mutable<Option<Rc<wellen::Hierarchy>>>,
    selected_var_refs: MutableVec<wellen::VarRef>,
    layout: Mutable<Layout>,
    loaded_filename: Mutable<Option<Filename>>,
}

impl ControlsPanel {
    pub fn new(
        hierarchy: Mutable<Option<Rc<wellen::Hierarchy>>>,
        selected_var_refs: MutableVec<wellen::VarRef>,
        layout: Mutable<Layout>,
    ) -> impl Element {
        Self {
            selected_scope_ref: <_>::default(),
            hierarchy,
            selected_var_refs,
            layout,
            loaded_filename: <_>::default(),
        }
        .root()
    }

    fn triggers(&self) -> Vec<TaskHandle> {
        vec![Task::start_droppable(clone!((self => s) async move {
            let was_some = Cell::new(false);
            s.hierarchy
                .signal_ref(Option::is_some)
                .dedupe()
                .for_each_sync(clone!((s) move |is_some| {
                    if is_some {
                        return was_some.set(true);
                    }
                    if was_some.get() {
                        s.selected_scope_ref.set(None);
                        s.selected_var_refs.lock_mut().clear();
                        s.loaded_filename.set(None);
                    }
                })).await
        }))]
    }

    fn root(&self) -> impl Element {
        let triggers = self.triggers();
        let layout = self.layout.clone();
        let layout_and_hierarchy_signal = map_ref! {
            let layout = layout.signal(),
            let hierarchy = self.hierarchy.signal_cloned() => {
                (*layout, hierarchy.clone())
            }
        };
        Column::new()
            .after_remove(move |_| drop(triggers))
            .s(Width::with_signal_self(
                self.layout
                    .signal()
                    .map(|layout| matches!(layout, Layout::Columns))
                    .map_true(|| Width::fill()),
            ))
            .s(Height::with_signal_self(layout.signal().map(
                move |layout| match layout {
                    Layout::Tree => Height::fill(),
                    Layout::Columns => Height::fill().max(MILLER_COLUMN_MAX_HEIGHT),
                },
            )))
            .s(Padding::all(20))
            .s(Gap::new().y(40))
            .s(Align::new().top())
            .item(
                Row::new()
                    .s(Gap::both(15))
                    .s(Align::new().left())
                    .item(self.load_button())
                    .item(self.layout_switcher()),
            )
            .item(self.command_panel())
            .item_signal(
                self.hierarchy
                    .signal_cloned()
                    .map_some(clone!((self => s) move |hierarchy| s.scopes_panel(hierarchy))),
            )
            .item_signal(layout_and_hierarchy_signal.map(
                clone!((self => s) move |(layout, hierarchy)| {
                    hierarchy.and_then(clone!((s) move |hierarchy| {
                        matches!(layout, Layout::Tree).then(move || s.vars_panel(hierarchy))
                    }))
                }),
            ))
    }

    #[cfg(FASTWAVE_PLATFORM = "TAURI")]
    fn load_button(&self) -> impl Element {
        let (hovered, hovered_signal) = Mutable::new_and_signal(false);
        let hierarchy = self.hierarchy.clone();
        let loaded_filename = self.loaded_filename.clone();
        Button::new()
            .s(Padding::new().x(20).y(10))
            .s(Background::new().color_signal(
                hovered_signal.map_bool(|| color!("MediumSlateBlue"), || color!("SlateBlue")),
            ))
            .s(Align::new().left())
            .s(RoundedCorners::all(15))
            .label(El::new().s(Font::new().no_wrap()).child_signal(
                loaded_filename.signal_cloned().map_option(
                    |filename| format!("Unload {filename}"),
                    || format!("Load file.."),
                ),
            ))
            .on_hovered_change(move |is_hovered| hovered.set_neq(is_hovered))
            .on_press(move || {
                let mut hierarchy_lock = hierarchy.lock_mut();
                if hierarchy_lock.is_some() {
                    *hierarchy_lock = None;
                    return;
                }
                drop(hierarchy_lock);
                let hierarchy = hierarchy.clone();
                let loaded_filename = loaded_filename.clone();
                Task::start(async move {
                    if let Some(filename) = platform::pick_and_load_waveform(None).await {
                        loaded_filename.set_neq(Some(filename));
                        hierarchy.set(Some(Rc::new(platform::get_hierarchy().await)))
                    }
                })
            })
    }

    #[cfg(FASTWAVE_PLATFORM = "BROWSER")]
    fn load_button(&self) -> impl Element {
        let (hovered, hovered_signal) = Mutable::new_and_signal(false);
        let hierarchy = self.hierarchy.clone();
        let loaded_filename = self.loaded_filename.clone();
        let file_input_id = "file_input";
        Row::new()
            .item(
                Label::new()
                    .s(Padding::new().x(20).y(10))
                    .s(Background::new().color_signal(
                        hovered_signal
                            .map_bool(|| color!("MediumSlateBlue"), || color!("SlateBlue")),
                    ))
                    .s(Align::new().left())
                    .s(RoundedCorners::all(15))
                    .s(Cursor::new(CursorIcon::Pointer))
                    .label(El::new().s(Font::new().no_wrap()).child_signal(
                        loaded_filename.signal_cloned().map_option(
                            |filename| format!("Unload {filename}"),
                            || format!("Load file.."),
                        ),
                    ))
                    .on_hovered_change(move |is_hovered| hovered.set_neq(is_hovered))
                    .for_input(file_input_id)
                    .on_click_event_with_options(
                        EventOptions::new().preventable(),
                        clone!((hierarchy) move |event| {
                            let mut hierarchy_lock = hierarchy.lock_mut();
                            if hierarchy_lock.is_some() {
                                *hierarchy_lock = None;
                                if let RawMouseEvent::Click(raw_event) = event.raw_event {
                                    // @TODO Move to MoonZoon as a new API
                                    raw_event.prevent_default();
                                }
                                return;
                            }
                        }),
                    ),
            )
            .item(
                // @TODO https://github.com/MoonZoon/MoonZoon/issues/39
                // + https://developer.mozilla.org/en-US/docs/Web/API/File_API/Using_files_from_web_applications#using_hidden_file_input_elements_using_the_click_method
                TextInput::new().id(file_input_id).update_raw_el(|raw_el| {
                    let dom_element = raw_el.dom_element();
                    raw_el
                        .style("display", "none")
                        .attr("type", "file")
                        .event_handler(move |_: events::Input| {
                            let Some(file_list) =
                                dom_element.files().map(gloo_file::FileList::from)
                            else {
                                zoon::println!("file list is `None`");
                                return;
                            };
                            let Some(file) = file_list.first().cloned() else {
                                zoon::println!("file list is empty");
                                return;
                            };
                            let hierarchy = hierarchy.clone();
                            let loaded_filename = loaded_filename.clone();
                            Task::start(async move {
                                if let Some(filename) =
                                    platform::pick_and_load_waveform(Some(file)).await
                                {
                                    loaded_filename.set_neq(Some(filename));
                                    hierarchy.set(Some(Rc::new(platform::get_hierarchy().await)))
                                }
                            })
                        })
                }),
            )
    }

    fn layout_switcher(&self) -> impl Element {
        let layout = self.layout.clone();
        let (hovered, hovered_signal) = Mutable::new_and_signal(false);
        Button::new()
            .s(Padding::new().x(20).y(10))
            .s(Background::new().color_signal(
                hovered_signal.map_bool(|| color!("MediumSlateBlue"), || color!("SlateBlue")),
            ))
            .s(Align::new().left())
            .s(RoundedCorners::all(15))
            .label_signal(layout.signal().map(|layout| match layout {
                Layout::Tree => "Columns",
                Layout::Columns => "Tree",
            }))
            .on_hovered_change(move |is_hovered| hovered.set_neq(is_hovered))
            .on_press(move || {
                layout.update(|layout| match layout {
                    Layout::Tree => Layout::Columns,
                    Layout::Columns => Layout::Tree,
                })
            })
    }

    fn command_panel(&self) -> impl Element {
        let command_result: Mutable<Option<Result<JsValue, JsValue>>> = <_>::default();
        Row::new()
            .item(self.command_editor_panel(command_result.clone()))
            .item(self.command_result_panel(command_result.read_only()))
    }

    fn command_editor_panel(
        &self,
        command_result: Mutable<Option<Result<JsValue, JsValue>>>,
    ) -> impl Element {
        let (script, script_signal) = Mutable::new_and_signal_cloned(String::new());
        TextArea::new()
            .placeholder(Placeholder::new("FW.say_hello()"))
            .label_hidden("command editor panel")
            .text_signal(script_signal)
            .on_change(clone!((script, command_result) move |text| {
                script.set_neq(text);
                command_result.set_neq(None);
            }))
            .on_key_down_event_with_options(EventOptions::new().preventable(), move |event| {
                if event.key() == &Key::Enter {
                    let RawKeyboardEvent::KeyDown(raw_event) = event.raw_event.clone();
                    if raw_event.shift_key() {
                        // @TODO move `prevent_default` to MZ API (next to the `pass_to_parent` method?)
                        raw_event.prevent_default();
                        let result = script_bridge::strict_eval(&script.lock_ref());
                        command_result.set(Some(result));
                    }
                }
            })
    }

    fn command_result_panel(
        &self,
        command_result: ReadOnlyMutable<Option<Result<JsValue, JsValue>>>,
    ) -> impl Element {
        El::new().child_signal(
            command_result
                .signal_cloned()
                .map_some(|result| match result {
                    Ok(js_value) => {
                        if let Some(string_value) = js_value.as_string() {
                            string_value
                        } else if let Some(number_value) = js_value.as_f64() {
                            number_value.to_string()
                        } else if let Some(bool_value) = js_value.as_bool() {
                            bool_value.to_string()
                        } else {
                            format!("{js_value:?}")
                        }
                    }
                    Err(js_value) => {
                        format!("Error: {js_value:?}")
                    }
                }),
        )
    }

    fn scopes_panel(&self, hierarchy: Rc<wellen::Hierarchy>) -> impl Element {
        Column::new()
            .s(Height::fill().min(150))
            .s(Scrollbars::y_and_clip_x())
            .s(Gap::new().y(20))
            .s(Width::fill())
            .item_signal(
                self.layout
                    .signal()
                    .map(|layout| matches!(layout, Layout::Tree))
                    .map_true(|| El::new().child("Scopes")),
            )
            .item(self.scopes_list(hierarchy))
    }

    fn scopes_list(&self, hierarchy: Rc<wellen::Hierarchy>) -> impl Element {
        let layout = self.layout.clone();
        let mut scopes_for_ui = Vec::new();
        let mut max_level_index: usize = 0;
        for scope_ref in hierarchy.scopes() {
            let mut scope_refs = Vec::new();
            scope_refs.push((0, scope_ref, None));
            let mut selected_scope_in_levels: Vec<Mutable<Option<wellen::ScopeRef>>> =
                vec![<_>::default()];
            while let Some((level, scope_ref, parent_expanded)) = scope_refs.pop() {
                let scope = hierarchy.get(scope_ref);
                let mut children = scope.scopes(&hierarchy).peekable();
                let has_children = children.peek().is_some();
                let expanded = Mutable::new(false);
                if level > max_level_index {
                    max_level_index = level;
                    selected_scope_in_levels.push(<_>::default());
                }
                scopes_for_ui.push(ScopeForUI {
                    level,
                    name: Rc::new(scope.name(&hierarchy).to_owned()),
                    scope_ref,
                    has_children,
                    expanded: expanded.clone(),
                    parent_expanded,
                    selected_scope_in_level: selected_scope_in_levels[level].clone(),
                });
                for scope_ref in children {
                    scope_refs.push((level + 1, scope_ref, Some(expanded.read_only())));
                }
            }
        }
        let scopes_for_ui = Rc::new(scopes_for_ui);
        let s = self.clone();
        El::new()
            .s(Height::fill())
            .s(Scrollbars::both())
            .s(Width::fill())
            .child_signal(layout.signal().map(move |layout| match layout {
                Layout::Tree => {
                    Column::new()
                        .s(Align::new().left())
                        .s(Gap::new().y(10))
                        .s(Height::fill())
                        .s(Scrollbars::y_and_clip_x())
                        .s(Padding::new().right(15))
                        .items(
                            scopes_for_ui
                                .iter()
                                .map(clone!((s) move |scope_for_ui| s.scope_button_row(scope_for_ui.clone()))),
                        ).unify()
                }
                Layout::Columns => {
                    let mut scopes_for_ui_in_levels: Vec<Vec<ScopeForUI>> = vec![Vec::new(); max_level_index + 1];
                    for scope_for_ui in scopes_for_ui.iter() {
                        scopes_for_ui_in_levels[scope_for_ui.level].push(scope_for_ui.clone());
                    }
                    let viewport_x = Mutable::new(0);
                    El::new()
                        .s(Height::fill())
                        .s(Scrollbars::x_and_clip_y())
                        .s(Padding::new().bottom(15))
                        .s(Width::fill())
                        .viewport_x_signal(viewport_x.signal())
                        .child(
                            Row::new()
                                .s(Height::fill())
                                // @TODO add `width: max-content` to MoonZoon's `Width`?
                                .update_raw_el(|raw_el| raw_el.style("width", "max-content"))
                                .on_viewport_size_change(move |_, _| viewport_x.set(i32::MAX))
                                .items(scopes_for_ui_in_levels.into_iter().map(|scopes_in_level| {
                                    Column::new()
                                        .s(Height::fill())
                                        .s(Scrollbars::y_and_clip_x())
                                        // @TODO `Width::default` add the class `exact_width` with `flex-shrink: 0;`
                                        // We should make it more explicit / discoverable in MoonZoon
                                        .s(Width::default())
                                        .s(Gap::new().y(10))
                                        .s(Padding::new().x(10))
                                        .items(
                                            scopes_in_level
                                                .into_iter()
                                                .map(clone!((s) move |scope_for_ui| s.scope_button_row(scope_for_ui)))
                                        )
                                }))
                                .item(s.vars_panel(hierarchy.clone()))
                        ).unify()
                }
            }))
    }

    fn scope_button_row(&self, scope_for_ui: ScopeForUI) -> impl Element {
        let layout = self.layout.clone();
        let (button_hovered, button_hovered_signal) = Mutable::new_and_signal(false);
        let selected_scope_ref = self.selected_scope_ref.clone();
        let is_selected = selected_scope_ref
            .signal()
            .map(move |selected_scope_ref| selected_scope_ref == Some(scope_for_ui.scope_ref));
        let background_color = map_ref! {
            let is_selected = is_selected,
            let is_hovered = button_hovered_signal => match (*is_selected, *is_hovered) {
                (true, _) => color!("BlueViolet"),
                (false, true) => color!("MediumSlateBlue"),
                (false, false) => color!("SlateBlue"),
            }
        };
        let task_collapse_on_parent_collapse = {
            let expanded = scope_for_ui.expanded.clone();
            scope_for_ui.parent_expanded.clone().map(|parent_expanded| {
                Task::start_droppable(parent_expanded.signal().for_each_sync(
                    move |parent_expanded| {
                        if not(parent_expanded) {
                            expanded.set_neq(false);
                        }
                    },
                ))
            })
        };
        let task_expand_or_collapse_on_selected_scope_in_level_change = {
            let expanded = scope_for_ui.expanded.clone();
            let scope_ref = scope_for_ui.scope_ref;
            let layout = layout.clone();
            Task::start_droppable(scope_for_ui.selected_scope_in_level.signal().for_each_sync(
                move |selected_scope_in_level| {
                    if matches!(layout.get(), Layout::Columns) {
                        if let Some(selected_scope) = selected_scope_in_level {
                            if selected_scope == scope_ref {
                                return expanded.set(true);
                            }
                        }
                        expanded.set(false);
                    }
                },
            ))
        };
        let display = signal::option(
            scope_for_ui
                .parent_expanded
                .clone()
                .map(|parent_expanded| parent_expanded.signal().map_false(|| "none")),
        )
        .map(Option::flatten);
        let level = scope_for_ui.level as u32;
        El::new()
            // @TODO Add `Display` Style to MoonZoon? Merge with `Visible` Style?
            .update_raw_el(|raw_el| raw_el.style_signal("display", display))
            .s(Padding::new().left_signal(layout.signal().map(move |layout| match layout {
                Layout::Tree => level * 30,
                Layout::Columns => 0,
            })))
            .s(Width::default().max(SCOPE_VAR_ROW_MAX_WIDTH))
            .after_remove(move |_| {
                drop(task_collapse_on_parent_collapse);
                drop(task_expand_or_collapse_on_selected_scope_in_level_change);
            })
            .child(
                Row::new()
                    .s(Background::new().color_signal(background_color))
                    .s(RoundedCorners::all(15))
                    .s(Clip::both())
                    .s(Align::new().left())
                    .items_signal_vec(layout.signal().map(clone!((self => s, scope_for_ui) move |layout| {
                        let toggle = scope_for_ui.has_children.then(clone!((s, scope_for_ui) move || s.scope_toggle(scope_for_ui)));
                        let button = s.scope_button(scope_for_ui.clone(), button_hovered.clone());
                        match layout {
                            Layout::Tree => element_vec![toggle, button],
                            Layout::Columns => element_vec![button, toggle],
                        }
                    })).to_signal_vec())
            )
    }

    fn scope_toggle(&self, scope_for_ui: ScopeForUI) -> impl Element {
        let layout = self.layout.clone();
        let expanded = scope_for_ui.expanded.clone();
        let layout_and_expanded = map_ref! {
            let layout = layout.signal(),
            let expanded = expanded.signal() => (*layout, *expanded)
        };
        let selected_scope_ref: Mutable<Option<wellen::ScopeRef>> = self.selected_scope_ref.clone();
        let (hovered, hovered_signal) = Mutable::new_and_signal(false);
        Button::new()
            .s(Padding::new()
                .left_signal(
                    self.layout
                        .signal()
                        .map(|layout| matches!(layout, Layout::Tree))
                        .map_true(|| 10),
                )
                .right_signal(
                    self.layout
                        .signal()
                        .map(|layout| matches!(layout, Layout::Columns))
                        .map_true(|| 10),
                ))
            .s(Height::fill())
            .s(Font::new().color_signal(hovered_signal.map_true(|| color!("LightBlue"))))
            .label(
                El::new()
                    .s(Transform::with_signal_self(layout_and_expanded.map(
                        |(layout, expanded)| match layout {
                            Layout::Tree => expanded.not().then(|| Transform::new().rotate(-90)),
                            Layout::Columns => {
                                Some(Transform::new().rotate(if expanded { -90 } else { 90 }))
                            }
                        },
                    )))
                    .child("▼"),
            )
            .on_hovered_change(move |is_hovered| hovered.set_neq(is_hovered))
            .on_press(move || match layout.get() {
                Layout::Tree => {
                    if scope_for_ui.expanded.get() {
                        scope_for_ui.selected_scope_in_level.set(None);
                    } else {
                        scope_for_ui
                            .selected_scope_in_level
                            .set(Some(scope_for_ui.scope_ref));
                    }
                    scope_for_ui.expanded.update(not)
                }
                Layout::Columns => {
                    selected_scope_ref.set_neq(None);
                    if scope_for_ui.expanded.get() {
                        scope_for_ui.selected_scope_in_level.set(None);
                    } else {
                        scope_for_ui
                            .selected_scope_in_level
                            .set(Some(scope_for_ui.scope_ref));
                    }
                }
            })
    }

    fn scope_button(
        &self,
        scope_for_ui: ScopeForUI,
        button_hovered: Mutable<bool>,
    ) -> impl Element {
        Button::new()
            .s(Padding::new().x(15).y(5))
            .s(Font::new().wrap_anywhere())
            .on_hovered_change(move |is_hovered| button_hovered.set_neq(is_hovered))
            .on_press(
                clone!((self.selected_scope_ref => selected_scope_ref, scope_for_ui) move || {
                    selected_scope_ref.set_neq(Some(scope_for_ui.scope_ref));
                    scope_for_ui.selected_scope_in_level.set_neq(None);
                }),
            )
            .label(scope_for_ui.name)
    }

    fn vars_panel(&self, hierarchy: Rc<wellen::Hierarchy>) -> impl Element {
        let selected_scope_ref = self.selected_scope_ref.clone();
        Column::new()
            .s(Gap::new().y(20))
            .s(Height::fill().min(150))
            .s(Scrollbars::y_and_clip_x())
            .item_signal(
                self.layout
                    .signal()
                    .map(|layout| matches!(layout, Layout::Tree))
                    .map_true(|| El::new().child("Variables")),
            )
            .item_signal(selected_scope_ref.signal().map_some(
                clone!((self => s) move |scope_ref| s.vars_list(scope_ref, hierarchy.clone())),
            ))
    }

    // @TODO Group variables?
    fn vars_list(
        &self,
        selected_scope_ref: wellen::ScopeRef,
        hierarchy: Rc<wellen::Hierarchy>,
    ) -> impl Element {
        let vars_for_ui = hierarchy
            .get(selected_scope_ref)
            .vars(&hierarchy)
            .map(|var_ref| {
                let var = hierarchy.get(var_ref);
                VarForUI {
                    name: Rc::new(var.name(&hierarchy).to_owned()),
                    var_type: var.var_type(),
                    var_direction: var.direction(),
                    var_ref,
                    signal_type: var.signal_tpe(),
                }
            });

        // Lazy loading to not freeze the main thread
        // @TODO replace with grouping and/or virtual scroll (https://dev.to/adamklein/build-your-own-virtual-scroll-part-i-11ib)
        const CHUNK_SIZE: usize = 50;
        let mut chunked_vars_for_ui: Vec<Vec<VarForUI>> = <_>::default();
        let mut chunk = Vec::with_capacity(CHUNK_SIZE);
        for (index, var_for_ui) in vars_for_ui.enumerate() {
            chunk.push(var_for_ui);
            if index % CHUNK_SIZE == 0 {
                chunked_vars_for_ui.push(mem::take(&mut chunk));
            }
        }
        if not(chunk.is_empty()) {
            chunked_vars_for_ui.push(chunk);
        }
        let vars_for_ui_mutable_vec = MutableVec::<VarForUI>::new();
        let append_vars_for_ui_task =
            Task::start_droppable(clone!((vars_for_ui_mutable_vec) async move {
                for chunk in chunked_vars_for_ui {
                    Task::next_macro_tick().await;
                    vars_for_ui_mutable_vec.lock_mut().extend(chunk);
                }
            }));

        Column::new()
            .s(Width::with_signal_self(
                self.layout
                    .signal()
                    .map(|layout| matches!(layout, Layout::Columns))
                    .map_true(|| Width::default().min(SCOPE_VAR_ROW_MAX_WIDTH)),
            ))
            .s(Align::new().left())
            .s(Gap::new().y(10))
            .s(Height::fill())
            .s(Scrollbars::y_and_clip_x())
            .items_signal_vec(
                vars_for_ui_mutable_vec
                    .signal_vec_cloned()
                    .map(clone!((self => s) move |var_for_ui| s.var_row(var_for_ui))),
            )
            .after_remove(move |_| drop(append_vars_for_ui_task))
    }

    fn var_row(&self, var_for_ui: VarForUI) -> impl Element {
        Row::new()
            .s(Gap::new().x(10))
            .s(Padding::new().right(15))
            .s(Width::default().max(SCOPE_VAR_ROW_MAX_WIDTH))
            .item(self.var_button(var_for_ui.clone()))
            .item(self.var_tag_type(var_for_ui.clone()))
            .item(self.var_tag_index(var_for_ui.clone()))
            .item(self.var_tag_bit(var_for_ui.clone()))
            .item(self.var_tag_direction(var_for_ui))
    }

    fn var_button(&self, var_for_ui: VarForUI) -> impl Element {
        let (hovered, hovered_signal) = Mutable::new_and_signal(false);
        let selected_var_ref = self.selected_var_refs.clone();
        El::new().child(
            Button::new()
                .s(Font::new().wrap_anywhere())
                .s(Padding::new().x(15).y(5))
                .s(Background::new().color_signal(
                    hovered_signal.map_bool(|| color!("MediumSlateBlue"), || color!("SlateBlue")),
                ))
                .s(RoundedCorners::all(15))
                .on_hovered_change(move |is_hovered| hovered.set_neq(is_hovered))
                .on_press(move || selected_var_ref.lock_mut().push(var_for_ui.var_ref))
                .label(var_for_ui.name),
        )
    }

    fn var_tag_type(&self, var_for_ui: VarForUI) -> impl Element {
        let var_type = var_for_ui.var_type;
        El::new().child(format!("{var_type:?}"))
    }

    fn var_tag_index(&self, var_for_ui: VarForUI) -> Option<impl Element> {
        let wellen::SignalType::BitVector(_, Some(index)) = var_for_ui.signal_type else {
            None?
        };
        let msb = index.msb();
        let lsb = index.lsb();
        El::new().child(format!("[{msb}:{lsb}]")).apply(Some)
    }

    fn var_tag_bit(&self, var_for_ui: VarForUI) -> Option<impl Element> {
        let wellen::SignalType::BitVector(length, _) = var_for_ui.signal_type else {
            None?
        };
        El::new()
            .s(Font::new().no_wrap())
            .child(format!("{length}-bit"))
            .apply(Some)
    }

    fn var_tag_direction(&self, var_for_ui: VarForUI) -> impl Element {
        let direction = match var_for_ui.var_direction {
            wellen::VarDirection::Unknown => String::new(),
            direction => format!("{direction:?}"),
        };
        El::new().child(direction)
    }
}
